import unittest
from unittest import mock

from ZEO.runzeo import ZEOServer


class TestStorageServer:

    def __init__(self, fail_create_server):
        self.called = []
        if fail_create_server:
            raise RuntimeError()

    def close(self):
        self.called.append("close")


class TestZEOServer(ZEOServer):

    def __init__(self, fail_create_server=False, fail_loop_forever=False):
        ZEOServer.__init__(self, None)
        self.called = []
        self.fail_create_server = fail_create_server
        self.fail_loop_forever = fail_loop_forever

    def setup_default_logging(self):
        self.called.append("setup_default_logging")

    def check_socket(self):
        self.called.append("check_socket")

    def clear_socket(self):
        self.called.append("clear_socket")

    def make_pidfile(self):
        self.called.append("make_pidfile")

    def open_storages(self):
        self.called.append("open_storages")

    def setup_signals(self):
        self.called.append("setup_signals")

    def create_server(self):
        self.called.append("create_server")
        self.server = TestStorageServer(self.fail_create_server)

    def loop_forever(self):
        self.called.append("loop_forever")
        if self.fail_loop_forever:
            raise RuntimeError()

    def close_server(self):
        self.called.append("close_server")
        ZEOServer.close_server(self)

    def remove_pidfile(self):
        self.called.append("remove_pidfile")


class AttributeErrorTests(unittest.TestCase):

    def testFailCreateServer(self):
        #
        # Fix AttributeError: 'ZEOServer' object has no attribute
        # 'server' in ZEOServer.main
        #
        # Demonstrate the AttributeError
        zeo = TestZEOServer(fail_create_server=True)
        self.assertRaises(RuntimeError, zeo.main)


class CloseServerTests(unittest.TestCase):

    def testCallSequence(self):
        # The close_server hook is called after loop_forever
        # has returned
        zeo = TestZEOServer()
        zeo.main()
        self.assertEqual(zeo.called, [
            "setup_default_logging",
            "check_socket",
            "clear_socket",
            "make_pidfile",
            "open_storages",
            "setup_signals",
            "create_server",
            "loop_forever",
            "close_server",  # New
            "clear_socket",
            "remove_pidfile",
            ])
        # The default implementation closes the storage server
        self.assertEqual(hasattr(zeo, "server"), True)
        self.assertEqual(zeo.server.called, ["close"])

    def testFailLoopForever(self):
        # The close_server hook is called if loop_forever exits
        # with an exception
        zeo = TestZEOServer(fail_loop_forever=True)
        self.assertRaises(RuntimeError, zeo.main)
        self.assertEqual(zeo.called, [
            "setup_default_logging",
            "check_socket",
            "clear_socket",
            "make_pidfile",
            "open_storages",
            "setup_signals",
            "create_server",
            "loop_forever",
            "close_server",
            "clear_socket",
            "remove_pidfile",
            ])
        # The storage server has been closed
        self.assertEqual(hasattr(zeo, "server"), True)
        self.assertEqual(zeo.server.called, ["close"])

    def testFailCreateServer(self):
        # The close_server hook is called if create_server exits
        # with an exception
        zeo = TestZEOServer(fail_create_server=True)
        self.assertRaises(RuntimeError, zeo.main)
        self.assertEqual(zeo.called, [
            "setup_default_logging",
            "check_socket",
            "clear_socket",
            "make_pidfile",
            "open_storages",
            "setup_signals",
            "create_server",
            "close_server",
            "clear_socket",
            "remove_pidfile",
            ])
        # The server attribute is present but None
        self.assertEqual(hasattr(zeo, "server"), True)
        self.assertEqual(zeo.server, None)


@mock.patch('os.unlink')
class TestZEOServerSocket(unittest.TestCase):

    def _unlinked(self, unlink, options):
        server = ZEOServer(options)
        server.clear_socket()
        unlink.assert_called_once()

    def _not_unlinked(self, unlink, options):
        server = ZEOServer(options)
        server.clear_socket()
        unlink.assert_not_called()

    def test_clear_with_native_str(self, unlink):
        class Options:
            address = "a str that does not exist"
        self._unlinked(unlink, Options)

    def test_clear_with_unicode_str(self, unlink):
        class Options:
            address = "a str that does not exist"
        self._unlinked(unlink, Options)

    def test_clear_with_bytes(self, unlink):
        class Options:
            address = b'a byte str that does not exist'

        self._not_unlinked(unlink, Options)

    def test_clear_with_tuple(self, unlink):
        class Options:
            address = ('abc', 1)
        self._not_unlinked(unlink, Options)



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
import random
import unittest

from ZEO.TransactionBuffer import TransactionBuffer


def random_string(size):
    """Return a random string of size size."""
    lst = [chr(random.randrange(256)) for i in range(size)]
    return "".join(lst)


def new_store_data():
    """Return arbitrary data to use as argument to store() method."""
    return random_string(8), random_string(random.randrange(1000))


def store(tbuf, resolved=False):
    data = new_store_data()
    tbuf.store(*data)
    if resolved:
        tbuf.server_resolve(data[0])
    return data


class TransBufTests(unittest.TestCase):

    def checkTypicalUsage(self):
        tbuf = TransactionBuffer(0)
        store(tbuf)
        store(tbuf)
        for o in tbuf:
            pass
        tbuf.close()

    def checkOrderPreserved(self):
        tbuf = TransactionBuffer(0)
        data = []
        for i in range(10):
            data.append((store(tbuf), False))
            data.append((store(tbuf, True), True))

        for i, (oid, d, resolved) in enumerate(tbuf):
            self.assertEqual((oid, d), data[i][0])
            self.assertEqual(resolved, data[i][1])
        tbuf.close()


def test_suite():
    test_loader = unittest.TestLoader()
    test_loader.testMethodPrefix = 'check'
    return test_loader.loadTestsFromTestCase(TransBufTests)



"""SSL configuration support
"""
import os
import sys


def ssl_config(section, server):
    import ssl

    cafile = capath = None
    auth = section.authenticate
    if auth:
        if os.path.isdir(auth):
            capath = auth
        elif auth != 'DYNAMIC':
            cafile = auth

    context = ssl.create_default_context(
        ssl.Purpose.CLIENT_AUTH if server else ssl.Purpose.SERVER_AUTH,
        cafile=cafile, capath=capath)

    if not auth:
        assert not server
        context.load_default_certs()

    if section.certificate:
        password = section.password_function
        if password:
            module, name = password.rsplit('.', 1)
            module = __import__(module, globals(), locals(), ['*'], 0)
            password = getattr(module, name)
        context.load_cert_chain(section.certificate, section.key, password)

    context.verify_mode = ssl.CERT_REQUIRED

    context.verify_flags |= ssl.VERIFY_X509_STRICT | (
        context.cert_store_stats()['crl'] and ssl.VERIFY_CRL_CHECK_LEAF)

    if server:
        context.check_hostname = False
        return context

    context.check_hostname = section.check_hostname

    return context, section.server_hostname


def server_ssl(section):
    return ssl_config(section, True)


def client_ssl(section):
    return ssl_config(section, False)


class ClientStorageConfig:

    def __init__(self, config):
        self.config = config
        self.name = config.getSectionName()

    def open(self):
        from ZEO.ClientStorage import ClientStorage

        # config.server is a multikey of socket-connection-address values
        # where the value is a socket family, address tuple.
        config = self.config

        addresses = [server.address for server in config.server]
        options = {}
        if config.blob_cache_size is not None:
            options['blob_cache_size'] = config.blob_cache_size
        if config.blob_cache_size_check is not None:
            options['blob_cache_size_check'] = config.blob_cache_size_check
        if config.client_label is not None:
            options['client_label'] = config.client_label

        ssl = config.ssl
        if ssl:
            options['ssl'] = ssl[0]
            options['ssl_server_hostname'] = ssl[1]

        return ClientStorage(
            addresses,
            blob_dir=config.blob_dir,
            shared_blob_dir=config.shared_blob_dir,
            storage=config.storage,
            cache_size=config.cache_size,
            cache=config.cache_path,
            name=config.name,
            read_only=config.read_only,
            read_only_fallback=config.read_only_fallback,
            server_sync=config.server_sync,
            wait_timeout=config.wait_timeout,
            **options)



#! /usr/bin/env python
"""Update version numbers and release dates for the next release.

usage: release.py version date

version should be a string like "3.2.0c1"
date should be a string like "23-Sep-2003"

The following files are updated:
    - setup.py
    - NEWS.txt
    - docs/guide/zodb.tex
    - src/ZEO/__init__.py
    - src/ZEO/version.txt
    - src/ZODB/__init__.py
"""

import fileinput
import os
import re

# In file filename, replace the first occurrence of regexp pat with
# string repl.
def replace(filename, pat, repl):
    from sys import stderr as e # fileinput hijacks sys.stdout
    foundone = False
    for line in fileinput.input([filename], inplace=True, backup="~"):
        if foundone:
            print line,
        else:
            match = re.search(pat, line)
            if match is not None:
                foundone = True

                new = re.sub(pat, repl, line)
                print new,

                print >> e, "In %s, replaced:" % filename
                print >> e, "   ", repr(line)
                print >> e, "   ", repr(new)

            else:
                print line,

    if not foundone:
        print >> e, "*" * 60, "Oops!"
        print >> e, "    Failed to find %r in %r" % (pat, filename)

# Nothing in our codebase cares about ZEO/version.txt.  Jeremy said
# someone asked for it so that a shell script could read up the ZEO
# version easily.
# Before ZODB 3.4, the ZEO version was one smaller than the ZODB version;
# e.g., ZEO 2.2.7 shipped with ZODB 3.2.7.  Now ZEO and ZODB share their
# version number.
def write_zeoversion(path, version):
    f = open(path, "w")
    print >> f, version
    f.close()

def main(args):
    version, date = args

    replace("setup.py",
            r'^VERSION = "\S+"$',
            'VERSION = "%s"' % version)
    replace("src/ZODB/__init__.py",
            r'__version__ = "\S+"',
            '__version__ = "%s"' % version)
    replace("src/ZEO/__init__.py",
            r'version = "\S+"',
            'version = "%s"' % version)
    write_zeoversion("src/ZEO/version.txt", version)
    replace("NEWS.txt",
            r"^Release date: .*",
            "Release date: %s" % date)
    replace("docs/guide/zodb.tex",
            r"release{\S+}",
            "release{%s}" % version)
if __name__ == "__main__":
    import sys
    main(sys.argv[1:])



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Python versions compatiblity
"""
import platform
import sys

from zodbpickle.pickle import Unpickler as _Unpickler


PYPY = getattr(platform, 'python_implementation', lambda: None)() == 'PyPy'
WIN = sys.platform.startswith('win')


class Unpickler(_Unpickler):
    # Python 3 doesn't allow assignments to find_global,
    # instead, find_class can be overridden

    find_global = None

    def find_class(self, modulename, name):
        if self.find_global is None:
            return super().find_class(modulename, name)
        return self.find_global(modulename, name)



##############################################################################
#
# Copyright (c) 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Utilities for setting up the server environment."""

import os


def parentdir(p, n=1):
    """Return the ancestor of p from n levels up."""
    d = p
    while n:
        d = os.path.dirname(d)
        if not d or d == '.':
            d = os.getcwd()
        n -= 1
    return d


class Environment:
    """Determine location of the Data.fs & ZEO_SERVER.pid files.

    Pass the argv[0] used to start ZEO to the constructor.

    Use the zeo_pid and fs attributes to get the filenames.
    """

    def __init__(self, argv0):
        v = os.environ.get("INSTANCE_HOME")
        if v is None:
            # looking for a Zope/var directory assuming that this code
            # is installed in Zope/lib/python/ZEO
            p = parentdir(argv0, 4)
            if os.path.isdir(os.path.join(p, "var")):
                v = p
            else:
                v = os.getcwd()
        self.home = v
        self.var = os.path.join(v, "var")
        if not os.path.isdir(self.var):
            self.var = self.home

        pid = os.environ.get("ZEO_SERVER_PID")
        if pid is None:
            pid = os.path.join(self.var, "ZEO_SERVER.pid")

        self.zeo_pid = pid
        self.fs = os.path.join(self.var, "Data.fs")



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""ZEO iterator protocol tests."""

import gc

import transaction
from ZODB.Connection import TransactionMetaData

from ..asyncio.testing import AsyncRPC


class IterationTests:

    def _assertIteratorIdsEmpty(self):
        # Account for the need to run a GC collection
        # under non-refcounted implementations like PyPy
        # for storage._iterator_gc to fully do its job.
        # First, confirm that it ran
        self.assertTrue(self._storage._iterators._last_gc > 0)
        gc_enabled = gc.isenabled()
        # make sure there's no race conditions cleaning out the weak refs
        gc.disable()
        try:
            self.assertEqual(0, len(self._storage._iterator_ids))
        except AssertionError:
            # Ok, we have ids. That should also mean that the
            # weak dictionary has the same length.

            self.assertEqual(len(self._storage._iterators),
                             len(self._storage._iterator_ids))
            # Now if we do a collection and re-ask for iterator_gc
            # everything goes away as expected.
            gc.enable()
            gc.collect()
            gc.collect()  # sometimes PyPy needs it twice to clear weak refs

            self._storage._iterator_gc()

            self.assertEqual(len(self._storage._iterators),
                             len(self._storage._iterator_ids))
            self.assertEqual(0, len(self._storage._iterator_ids))
        finally:
            if gc_enabled:
                gc.enable()
            else:
                gc.disable()

    def checkIteratorGCProtocol(self):
        # Test garbage collection on protocol level.
        server = AsyncRPC(self._storage._server)

        iid = server.iterator_start(None, None)
        # None signals the end of iteration.
        self.assertEqual(None, server.iterator_next(iid))
        # The server has disposed the iterator already.
        self.assertRaises(KeyError, server.iterator_next, iid)

        iid = server.iterator_start(None, None)
        # This time, we tell the server to throw the iterator away.
        server.iterator_gc([iid])
        self.assertRaises(KeyError, server.iterator_next, iid)

    def checkIteratorExhaustionStorage(self):
        # Test the storage's garbage collection mechanism.
        self._dostore()
        iterator = self._storage.iterator()

        # At this point, a wrapping iterator might not have called the CS
        # iterator yet. We'll consume one item to make sure this happens.
        next(iterator)
        self.assertEqual(1, len(self._storage._iterator_ids))
        iid = list(self._storage._iterator_ids)[0]
        self.assertEqual([], list(iterator))
        self.assertEqual(0, len(self._storage._iterator_ids))

        # The iterator has run through, so the server has already disposed it.
        self.assertRaises(KeyError, self._storage._call, 'iterator_next', iid)

    def checkIteratorGCSpanTransactions(self):
        # Keep a hard reference to the iterator so it won't be automatically
        # garbage collected at the transaction boundary.
        self._dostore()
        iterator = self._storage.iterator()
        self._dostore()
        # As the iterator was not garbage collected, we can still use it. (We
        # don't see the transaction we just wrote being picked up, because
        # iterators only see the state from the point in time when they were
        # created.)
        self.assertTrue(list(iterator))

    def checkIteratorGCStorageCommitting(self):
        # We want the iterator to be garbage-collected, so we don't keep any
        # hard references to it. The storage tracks its ID, though.

        # The odd little jig we do below arises from the fact that the
        # CS iterator may not be constructed right away if the CS is wrapped.
        # We need to actually do some iteration to get the iterator created.
        # We do a store to make sure the iterator isn't exhausted right away.
        self._dostore()
        next(self._storage.iterator())

        self.assertEqual(1, len(self._storage._iterator_ids))
        iid = list(self._storage._iterator_ids)[0]

        # GC happens at the transaction boundary. After that, both the storage
        # and the server have forgotten the iterator.
        self._storage._iterators._last_gc = -1
        self._dostore()
        self._assertIteratorIdsEmpty()
        self.assertRaises(KeyError, self._storage._call, 'iterator_next', iid)

    def checkIteratorGCStorageTPCAborting(self):
        # Disabling GC to prevent a race condition in PyPy later on:
        gc.disable()
        # The odd little jig we do below arises from the fact that the
        # CS iterator may not be constructed right away if the CS is wrapped.
        # We need to actually do some iteration to get the iterator created.
        # We do a store to make sure the iterator isn't exhausted right away.
        self._dostore()
        next(self._storage.iterator())

        iid = list(self._storage._iterator_ids)[0]

        t = TransactionMetaData()
        self._storage._iterators._last_gc = -1
        self._storage.tpc_begin(t)
        self._storage.tpc_abort(t)
        self._assertIteratorIdsEmpty()
        self.assertRaises(KeyError, self._storage._call, 'iterator_next', iid)
        gc.enable()

    def checkIteratorGCStorageDisconnect(self):

        # The odd little jig we do below arises from the fact that the
        # CS iterator may not be constructed right away if the CS is wrapped.
        # We need to actually do some iteration to get the iterator created.
        # We do a store to make sure the iterator isn't exhausted right away.
        self._dostore()
        next(self._storage.iterator())

        t = TransactionMetaData()
        self._storage.tpc_begin(t)
        # Show that after disconnecting, the client side GCs the iterators
        # as well. I'm calling this directly to avoid accidentally
        # calling tpc_abort implicitly.
        self._storage.notify_disconnected()
        self.assertEqual(0, len(self._storage._iterator_ids))
        # maybe, ``notify_disconnected`` should automatically clean up
        self._storage.tpc_abort(t)  # avoid ``ResourceWarning``

    def checkIteratorParallel(self):
        self._dostore()
        self._dostore()
        iter1 = self._storage.iterator()
        iter2 = self._storage.iterator()
        txn_info1 = next(iter1)
        txn_info2 = next(iter2)
        self.assertEqual(txn_info1.tid, txn_info2.tid)
        txn_info1 = next(iter1)
        txn_info2 = next(iter2)
        self.assertEqual(txn_info1.tid, txn_info2.tid)
        self.assertRaises(StopIteration, next, iter1)
        self.assertRaises(StopIteration, next, iter2)


def iterator_sane_after_reconnect():
    r"""Make sure that iterators are invalidated on disconnect.

Start a server:

    >>> addr, adminaddr = start_server(  # NOQA: F821 undefined
    ...     '<filestorage>\npath fs\n</filestorage>', keep=1)

Open a client storage to it and commit a some transactions:

    >>> import ZEO, ZODB
    >>> client = ZEO.client(addr)
    >>> db = ZODB.DB(client)
    >>> conn = db.open()
    >>> for i in range(10):
    ...     conn.root().i = i
    ...     transaction.commit()

Create an iterator:

    >>> it = client.iterator()
    >>> tid1 = it.next().tid

Restart the storage:

    >>> stop_server(adminaddr)  # NOQA: F821 undefined
    >>> wait_disconnected(client)  # NOQA: F821 undefined
    >>> _ = start_server(  # NOQA: F821 undefined
    ...         '<filestorage>\npath fs\n</filestorage>', addr=addr)
    >>> wait_connected(client)  # NOQA: F821 undefined

Now, we'll create a second iterator:

    >>> it2 = client.iterator()

If we try to advance the first iterator, we should get an error:

    >>> it.next().tid > tid1
    Traceback (most recent call last):
    ...
    ClientDisconnected: Disconnected iterator

The second iterator should be peachy:

    >>> it2.next().tid == tid1
    True

Cleanup:

    >>> db.close()
    """
